;; This version for simulator - Uses approximate q in generators
;; Load flow caluculation in Cartesian coordinates - corrected equations
;; Each processor plays three roles:
;;    1) Calculates c, d and sends to partner at other end of wire
;;    2) Recieves c, d from partner and participates in sums for C, D
;;    3) Represents its node, calculates new u, v (redundantly)

(*proclaim '(type fixnum mxn mxe))

(defvar mxn 0)   ;; Maximum number of nodes = grid size in 0 coord
(defvar mxe 0)   ;; Maximum degree of nodes = grid size in 1 coord

;; Network representation;

(*proclaim
 '(type single-float-pvar p q u u2 v w gm bm cm dm C D f a
                          alfa beta gmma dlta Aq Bq Cq wk))

(*defvar p    (!! 0.0))    ; p for node
(*defvar q    (!! 0.0))    ; q for node
(*defvar u    (!! 0.0))    ; u for node (generators) u^2 = v^2 + w^2
(*defvar u2   (!! 0.0))    ; u^2 for node
(*defvar v    (!! 0.0))    ; v for node
(*defvar w    (!! 0.0))    ; w for node
(*defvar gm   (!! 0.0))    ; g for line
(*defvar bm   (!! 0.0))    ; b for line
(*defvar cm   (!! 0.0))    ; cm = gm*v - bm*w
(*defvar dm   (!! 0.0))    ; dm = gm*w + bm*v
(*defvar C    (!! 0.0))    ; C = SUM cm over adjacent nodes
(*defvar D    (!! 0.0))    ; D = SUM dm over adjacent nodes
(*defvar f    (!! 0.0))    ; 1/Ykk = f + j*a
(*defvar a    (!! 0.0))
(*defvar alfa (!! 0.0))    ; alfa = f*C - a*D
(*defvar beta (!! 0.0))    ; beta = f*D + a*C
(*defvar gmma (!! 0.0))    ; w = gmma*v + dlta in solution
(*defvar dlta (!! 0.0))    ; 
(*defvar Aq   (!! 0.0))    ; At every node, quadratic eqn is
(*defvar Bq   (!! 0.0))    ;    Aq*v^2 + Bq^v + Cq = 0
(*defvar Cq   (!! 0.0))    ; 
(*defvar wk   (!! 0.0))    ; permanent working variable

(*proclaim '(type (unsigned-pvar *current-send-address-length*) to))

(*defvar to (!! 0)) ;; Pointer to partner

(*proclaim '(type (unsigned-pvar 20) myx myy))

(*defvar myx (self-address-grid!! (!! 0))) ;; My node number
(*defvar myy (self-address-grid!! (!! 1))) ;; My position in the group


(*proclaim '(type Boolean-pvar act snd gen clc))

(*defvar act nil!!)        ;; Active in graph rep
(*defvar snd nil!!)        ;; I am next proc to use as sender for my node
(*defvar gen nil!!)        ;; I am a generator
(*defvar clc nil!!)        ;; I calculate new values (not a slack bus)

;; Set up CM with specified physical size, virtual dimensions from mxn mxe

(defun setup (sz)
 ;; (cm:attach sz) Removed for simulator
 (*cold-boot :initial-dimensions (list mxn mxe))
)

(defvar NODES 0)   ; Actual number of nodes
(defvar EDGES 0)   ; Actual number of edges (transmission lines)

;; Enter edge from na to nt with coefficients gx, bx

(defun enter-edge (na nt gx bx)
 (let ((rc 0))
      (declare (type fixnum rc))
   (incf EDGES)
   (when (and (*or (and!! (=!! myx (!! nt)) snd))
	      (*or (and!! (=!! myx (!! na))  snd))) ;; Receiver and sender OK
    (*when (and!! (=!! myx (!! nt)) snd)
      (do-for-selected-processors (c) ; exactly one processor
        (*setf (pref act c) t)  (setq rc c)
        (*setf (pref gm c) gx)   (*setf (pref bm c) bx)
 	(if (< (pref myy c) (1- mxe)) (*news snd snd 0 1))
        (*setf (pref snd c) nil))
     )  ; end *when
    (*when (and!! (=!! myx (!! na)) snd)
      (do-for-selected-processors (c) ; exactly one processor
        (*setf (pref act c) t)
	(*setf (pref to c) rc)  (*setf (pref to rc) c)
        (*setf (pref gm c) gx)   (*setf (pref bm c) bx)
	(if (< (pref myy c) (1- mxe)) (*news snd snd 0 1))
	(*setf (pref snd c) nil))
      )))  ; end *when, end when (and ...), end let
)  ; end enter-edge

;; Read graph description from stream and install - Cm should be setup
;; suitably.  Ignores edges which won't fit.

(defun loadg (strm)
 (*set gm (!! 0.0)) (*set bm (!! 0.0))
 (*set  act nil!!)  (*set snd nil!!)
 (*when (=!! myy (!! 0)) (*set snd t!!))
 (setq NODES (setq EDGES 0))
 (do ((n (read strm) (read strm)))
     ((< n 0))
   (incf NODES)
   (do ((nt (read strm) (read strm))
        (gx (read strm) (read strm)) (bx (read strm) (read strm)))
       ((< nt 0))
     (enter-edge n nt gx bx)
    )  ; end do nt, gx, bx
  )  ; end do n
 (list NODES EDGES)
)  ; end loadg

;; Load node data for generators

(defun loadgen (strm)
 (do ((n (read strm) (read strm)))
     ((< n 0))
  (let ((px (read strm)) (ux (read strm)))
    (*when (=!! myx (!! n))
       (*setf p (!! px)) (*setf u (!! ux)) (*set gen t!!)))
 ) ; end do
)

;; Load node data for loads

(defun loadld (strm)
 (do ((n (read strm) (read strm)))
     ((< n 0))
  (let ((px (read strm)) (qx (read strm)))
    (*when (=!! myx (!! n))
       (*setf p (!! px)) (*setf q (!! qx)) (*set gen nil!!)))
 ) ; end do
)

;; Initialize for v, w, derived quantities for obtaining solution

(defun initvw ()
  (*set w (!! 0.0))  (*set u2 (*!! u u))
  (*if gen (*set v u) (*set v (!! 1.0)))
  (*set C (!! 0))  (*set D (!! 0))
  (*set clc (and!! act (plusp!! myx)))
  (*when act
    (*pset :no-collisions gm f to)  ;; For calculation of 1/Ykk
    (*pset :no-collisions bm a to)
    (*set f (reduce-and-spread!! f '+!! 1))
    (*set a (reduce-and-spread!! a '+!! 1))
    (*set wk (+!! (*!! f f) (*!! a a)))
    (*set f (-!! (/!! f wk)))  (*set a (/!! a wk)))
)

;; Perform one iteration toward solution.  Returns maximum absolute
;; change in v, w components.
;; This version assumes that D is never 0 in a generator.

(*proclaim '(type single-float-pvar oldv oldw))

(*defvar oldv (!! 0.0))   ; For convergence test
(*defvar oldw (!! 0.0))

(*proclaim '(type single-float maxch))

(defvar maxch 0.0)        ; maximum change in last iteration

(defun iter ()
  (*set oldv v) (*set oldw w)
  (*when act
    (*set cm (-!! (*!! gm v) (*!! bm w)))
    (*set dm (+!! (*!! gm w) (*!! bm v)))
    (*pset :no-collisions cm C to)  (*pset :no-collisions dm D to)
    (*set C (reduce-and-spread!! C '+!! 1))
    (*set D (reduce-and-spread!! D '+!! 1))
    (*set alfa (-!! (*!! f C) (*!! a D)))
    (*set beta (+!! (*!! f D) (*!! a C)))
    (*if gen  ;; Estimate q
      (*set q (-!! (*!! v C) (*!! u D))))
    (*set gmma (/!! beta alfa))
    (*set dlta (/!! (-!! (*!! f q) (*!! a p)) alfa))
    (*set Aq (+!! (!! 1.0) (*!! gmma gmma)))
    (*set Bq (*!! (!! 2.0) gmma dlta))
    (*set Cq (-!! (+!! (*!! dlta dlta) (*!! beta dlta)) (*!! f p) (*!! a q)))
  ;; Solve quadratic equation
    (*set wk (-!! (*!! Bq Bq) (*!! (!! 4.0) Aq Cq)))
    (*set wk (sqrt!! wk))
    (*when clc  ;; All but slack node
      (*set v (/!! (-!! wk Bq) (*!! (!! 2.0) Aq)))
      (*set w (+!! (*!! gmma v) dlta))
      (*set wk (abs!! (-!! w oldw)))
      (setq maxch (*max wk))
      (*set wk (abs!! (-!! v oldv)))
      (setq maxch (max maxch (*max wk))))
  ) ; end *when act
  maxch
)

;; Double current graph with connections of length cl.  Intended for
;; situations in which nodes are 0, ..., NODES-1 (contiguous).
;; Not yet updated for networks.

(defun double (cl)
 (let ((sndn 0) (rcvn 0))
      (declare (type fixnum sndn rcvn))
   (*when snd (setq sndn (*min myx)))
   (*when rcv (setq rcvn (*min myx)))
   (when (or (null snd) (null rcv) (> (* 2 NODES) mxn))
         (format t "Can't double the graph.~%")
	 (return-from double))
   (*let ((buddy (!! 0)))
     (*if (<!! myx (!! NODES))
	  (*set buddy (cube-from-grid-address!! (+!! myx (!! NODES)) myy))
	  (*set buddy (cube-from-grid-address!! (-!! myx (!! NODES)) myy)))
     (*when (<!! myx (!! NODES))
       (*pset :no-collisions lng lng buddy :notify act) ; sets act
       (*pset :no-collisions snd snd buddy)
       (*pset :no-collisions rcv rcv buddy)
       (*pset :no-collisions (+!! fr (!! NODES)) fr buddy)
       (*pset :no-collisions (pref!! buddy to) to buddy))
     (setq EDGES (* 2 EDGES)) ;; Will be advanced by enter-edge
     (enter-edge sndn (+ rcvn NODES) cl)
     (enter-edge (+ sndn NODES) rcvn cl)
     (setq NODES (* 2 NODES))
   )  ; end *let
 (format t "The graph has ~D nodes, ~D edges.~%" NODES EDGES)
 )  ; end let
)  ; end double

(defvar file "gtest.dat")      ; The current data file (a string)

;; Load graph from data file whose name is in file.

(defun sg ()
  (with-open-file (strmx file) (loadg strmx) (loadgen strmx) (loadld strmx))
)
